import java.util.ArrayList;

/**
 * This class implements a set of integers: It stores a collection of 
 * integer values without allowing duplicates, and supports methods 
 * for adding or removing either individual values, or all values found 
 * in another set.
 * 
 * @author Brad Richards
 * @version 1.0
 */

public class IntSet {
    
    // Holds the items in the set. The list is kept ordered.
    private ArrayList<Integer> values;  

    /**
     * Constructor creates storage for future set elements.
     */
    public IntSet() {
        values = new ArrayList<Integer>();
    }

    /**
     * This private method is used as a "helper" in the add and remove methods below.
     * It does a binary search on the ArrayList looking for the specified key, and 
     * returns the final position (index) at which the search stops. This is the
     * position at which the key *will* be found if it's in the ArrayList, but it's
     * also the position at which key should be *inserted* (or close, anyway) if
     * it's not there.
     * 
     * @param key  An item whose position in the list is to be found.
     * @return  Position at which the binary search stops.
     */
    public int find(int key) {
        int left = 0;
        int right = values.size(); 
        while(right - left > 1) {
            // System.out.println("Searching "+left+".."+right);
            int mid = (left+right)/2;
            if (key < values.get(mid)) {
                right = mid;  // mid is now the right edge of search area
            }
            else {
                left = mid;
            }
        }
        return left;  // We narrowed it down to this position
    }

    /**
     * Returns a string displaying the set's contents.
     * @return String containing set's contents.
     */
    public String toString() {
        return values.toString();
    }

    /**
     * Returns the number of elements in the set.
     * @return  Number of elements in the set.
     */
    public int size() {
        return values.size();
    }

    /**
     * Returns true if the set contains the specified element, false otherwise.
     * @param key  The item we're looking for.
     * @return    Returns true if key is in the set.
     */
    public boolean contains(int key) {
        int location = find(key);
        return values.get(location) == key;  // Check "final box"
    }

    /**
     * Add a single value to the set, avoiding duplicates. 
     * @param key  The value to add.
     */
    public void add(int key) {
        if (values.size() > 0) {
            int location = find(key);  // Where it *would* be
            if (key > values.get(location)) {
                values.add(location+1, key);    // Add to the right
            }
            else if (key < values.get(location)) {
                values.add(location, key);      // Add in "final box"
            }
            // If key is ==, don't add it
        }
        else {
            values.add(key);
        }
    }

    /** 
     * Add in all items from the other IntSet.
     * @param other  Set containing items to add
     */
    public void merge(IntSet other) {
        for(int n : other.values) {
            add(n);
        }
    }

    /**
     * Remove a single value from the set, if it exists.
     * @param key  The value to remove.
     */
    public void remove(int key) {
        if (values.size() > 0) {
            int location = find(key);    // Where it would be
            if (values.get(location) == key) {
                values.remove(location); // Remove if it's there
            }
        }
    }

    /**
     * Subtract off items in the other IntSet from this IntSet.
     * @param other  Set containing items to remove
     */
    public void subtract(IntSet other) {
        for(int n : other.values) {
            remove(n);
        }
    }
}